@@ -1,17 +1,24 @@
+from typing import Set, TYPE_CHECKING
 import logging
+
 l = logging.getLogger("claripy.frontends.composite_frontend")
 
 import weakref
 import itertools
+
 symbolic_count = itertools.count()
 
 from .constrained_frontend import ConstrainedFrontend
 from claripy.ast.strings import String
 
+if TYPE_CHECKING:
+    from claripy import SolverCompositeChild
+
+
 class CompositeFrontend(ConstrainedFrontend):
     def __init__(self, template_frontend, template_frontend_string, track=False, **kwargs):
         super().__init__(**kwargs)
-        self._solvers = { }
+        self._solvers = {}
         self._unchecked_solvers = weakref.WeakSet()
         self._owned_solvers = weakref.WeakSet()
         self._template_frontend = template_frontend
@@ -23,9 +30,9 @@ class CompositeFrontend(ConstrainedFrontend):
         super()._blank_copy(c)
         c._unchecked_solvers = weakref.WeakSet()
         c._owned_solvers = weakref.WeakSet()
-        c._solvers = { }
+        c._solvers = {}
         c._template_frontend = self._template_frontend
-        if hasattr(self, '_template_frontend_string'):
+        if hasattr(self, "_template_frontend_string"):
             c._template_frontend_string = self._template_frontend_string
         c._unsat = False
         c._track = self._track
@@ -37,10 +44,9 @@ class CompositeFrontend(ConstrainedFrontend):
 
         c._solvers = dict(self._solvers)
         c._unchecked_solvers = weakref.WeakSet(self._unchecked_solvers)
-        self._owned_solvers = weakref.WeakSet() # for the COW
+        self._owned_solvers = weakref.WeakSet()  # for the COW
         return c
 
-
     #
     # Serialization stuff
     #
@@ -65,9 +71,10 @@ class CompositeFrontend(ConstrainedFrontend):
     @property
     def _solver_list(self):
         seen_solvers = set()
-        solver_list = [ ]
+        solver_list = []
         for s in self._solvers.values():
-            if id(s) in seen_solvers: continue
+            if id(s) in seen_solvers:
+                continue
             seen_solvers.add(id(s))
             solver_list.append(s)
         return solver_list
@@ -87,18 +94,20 @@ class CompositeFrontend(ConstrainedFrontend):
 
     def _solvers_for_variables(self, names):
         seen_solvers = set()
-        existing_solvers = [ ]
+        existing_solvers = []
         for n in names:
-            if n not in self._solvers: continue
+            if n not in self._solvers:
+                continue
             s = self._solvers[n]
 
-            if id(s) in seen_solvers: continue
+            if id(s) in seen_solvers:
+                continue
             seen_solvers.add(id(s))
             existing_solvers.append(s)
         return existing_solvers
 
     @staticmethod
-    def _names_for(names=None, lst=None, lst2=None, e=None, v=None):
+    def _names_for(names=None, lst=None, lst2=None, e=None, v=None) -> Set[str]:
         if names is None:
             names = set()
         if e is not None and isinstance(e, Base):
@@ -118,9 +127,33 @@ class CompositeFrontend(ConstrainedFrontend):
     def _merged_solver_for(self, *args, **kwargs):
         return self._solver_for_names(self._names_for(*args, **kwargs))
 
-    def _solver_for_names(self, names):
+    def _solver_for_names(self, names: Set[str]) -> "SolverCompositeChild":
+        """
+        Get a merged child solver for variables specified in `names`.
+
+        :param names:   A set of variable names.
+        :return:        A composite child solver.
+        """
+
         l.debug("composite_solver._merged_solver_for() running with %d names", len(names))
-        solvers = self._solvers_for_variables(names)
+
+        # compute a transitive closure for all variable names
+        all_names = set(names)
+        new_names = names
+        solvers = set()
+        while True:
+            tmp_solvers = self._solvers_for_variables(new_names)
+            solvers |= set(tmp_solvers)
+            all_names |= new_names
+            tmp_names = set()
+            for solver in solvers:
+                tmp_names |= solver.variables
+            new_names = tmp_names.difference(all_names)
+            if not new_names:
+                break
+
+        solvers = list(solvers)
+
         if len(solvers) == 0:
             if any(var for var in names if var.startswith(String.STRING_TYPE_IDENTIFIER)):
                 l.debug("... creating new solver for strings")
@@ -140,28 +173,30 @@ class CompositeFrontend(ConstrainedFrontend):
         Returns a sequence of the solvers that self and others share.
         """
 
-        solvers_by_id = { id(s): s for s in self._solver_list }
+        solvers_by_id = {id(s): s for s in self._solver_list}
         common_solvers = set(solvers_by_id.keys())
-        other_sets = [ { id(s) for s in cs._solver_list } for cs in others ]
-        for o in other_sets: common_solvers &= o
+        other_sets = [{id(s) for s in cs._solver_list} for cs in others]
+        for o in other_sets:
+            common_solvers &= o
 
-        return [ solvers_by_id[s] for s in common_solvers ]
+        return [solvers_by_id[s] for s in common_solvers]
 
     def _variable_sets(self):
-        return { s.variables for s in self._solver_list }
+        return {s.variables for s in self._solver_list}
 
     def _shared_varsets(self, others):
         common_varsets = self._variable_sets()
-        for o in others: common_varsets &= o.all_varsets()
+        for o in others:
+            common_varsets &= o.all_varsets()
         return common_varsets
 
     def _split_child(self, s):
         ss = s.split()
         if len(ss) == 1:
-            return [ s ]
+            return [s]
 
         l.debug("... split solver %r into %d parts", s, len(ss))
-        l.debug("... variable counts: %s", [ len(cs.variables) for cs in ss ])
+        l.debug("... variable counts: %s", [len(cs.variables) for cs in ss])
 
         for ns in ss:
             self._owned_solvers.add(ns)
@@ -195,12 +230,12 @@ class CompositeFrontend(ConstrainedFrontend):
 
     def _store_child(self, ns, extra_names=frozenset(), invalidate_cache=True):
         for v in ns.variables | extra_names:
-            #os = self._solvers[v]
+            # os = self._solvers[v]
             self._solvers[v] = ns
         if invalidate_cache:
             self._unchecked_solvers.add(ns)
 
-        #if isinstance(s, ModelCacheMixin):
+        # if isinstance(s, ModelCacheMixin):
         #   if len(os._models) < len(ns._models):
         #       print("GOT %d NEW MODELS (before: %d)" % (
         #           len(ns._models) - len(os._models), len(os._models)
@@ -227,7 +262,7 @@ class CompositeFrontend(ConstrainedFrontend):
     def _add_dependent_constraints(self, names, constraints, invalidate_cache=True, **kwargs):
         if not invalidate_cache and len(self._solvers_for_variables(names)) > 1:
             l.debug("Ignoring cross-solver helper constraints.")
-            return [ ]
+            return []
 
         l.debug("Adding %d constraints to %d names", len(constraints), len(names))
         s = self._claim(self._merged_solver_for(names=names))
@@ -235,14 +270,14 @@ class CompositeFrontend(ConstrainedFrontend):
         self._store_child(s, invalidate_cache=invalidate_cache)
         return added
 
-    def add(self, constraints, **kwargs): #pylint:disable=arguments-differ
+    def add(self, constraints, **kwargs):  # pylint:disable=arguments-differ
         split = self._split_constraints(constraints)
-        child_added = [ ]
+        child_added = []
 
-        #l.debug("%s, solvers before: %d", self, len(self._solvers))
-        unsure = [ ]
-        for names,set_constraints in split:
-            if names == { 'CONCRETE' }:
+        # l.debug("%s, solvers before: %d", self, len(self._solvers))
+        unsure = []
+        for names, set_constraints in split:
+            if names == {"CONCRETE"}:
                 try:
                     if any(backends.concrete.convert(c) is False for c in set_constraints):
                         self._unsat = True
@@ -251,7 +286,7 @@ class CompositeFrontend(ConstrainedFrontend):
             else:
                 child_added += self._add_dependent_constraints(names, set_constraints, **kwargs)
 
-        #l.debug("... solvers after add: %d", len(self._solver_list))
+        # l.debug("... solvers after add: %d", len(self._solver_list))
 
         if len(unsure) > 0:
             for s in self._solver_list:
@@ -271,14 +306,14 @@ class CompositeFrontend(ConstrainedFrontend):
 
     def check_satisfiability(self, extra_constraints=(), exact=None):
         if self._unsat:
-            return 'UNSAT'
+            return "UNSAT"
 
         l.debug("%r checking satisfiability...", self)
 
         if len(extra_constraints) != 0:
             extra_solver = self._merged_solver_for(lst=extra_constraints)
             extra_solver_satness = extra_solver.check_satisfiability(extra_constraints=extra_constraints, exact=exact)
-            if extra_solver_satness in {'UNSAT', 'UNKNOWN'}:
+            if extra_solver_satness in {"UNSAT", "UNKNOWN"}:
                 return extra_solver_satness
             self._reabsorb_solver(extra_solver)
 
@@ -293,14 +328,14 @@ class CompositeFrontend(ConstrainedFrontend):
                 continue
 
             satness = s.check_satisfiability(exact=exact)
-            if satness in {'UNSAT', 'UNKNOWN'}:
+            if satness in {"UNSAT", "UNKNOWN"}:
                 return satness
 
         self._unchecked_solvers.clear()
-        return 'SAT'
+        return "SAT"
 
     def satisfiable(self, extra_constraints=(), exact=None):
-        return self.check_satisfiability(extra_constraints=extra_constraints, exact=exact) == 'SAT'
+        return self.check_satisfiability(extra_constraints=extra_constraints, exact=exact) == "SAT"
 
     def eval(self, e, n, extra_constraints=(), exact=None):
         self._ensure_sat(extra_constraints=extra_constraints)
@@ -343,26 +378,26 @@ class CompositeFrontend(ConstrainedFrontend):
         return r
 
     def is_true(self, e, extra_constraints=(), exact=None):
-        #self._ensure_sat(extra_constraints=extra_constraints)
+        # self._ensure_sat(extra_constraints=extra_constraints)
 
         ms = self._merged_solver_for(e=e, lst=extra_constraints)
         r = ms.is_true(e, extra_constraints=extra_constraints, exact=exact)
-        #self._reabsorb_solver(ms)
+        # self._reabsorb_solver(ms)
         return r
 
     def is_false(self, e, extra_constraints=(), exact=None):
-        #self._ensure_sat(extra_constraints=extra_constraints)
+        # self._ensure_sat(extra_constraints=extra_constraints)
 
         ms = self._merged_solver_for(e=e, lst=extra_constraints)
         r = ms.is_false(e, extra_constraints=extra_constraints, exact=exact)
-        #self._reabsorb_solver(ms)
+        # self._reabsorb_solver(ms)
         return r
 
     def unsat_core(self, extra_constraints=()):
         if self.satisfiable(extra_constraints=extra_constraints):
-            return tuple()
+            return ()
 
-        cores = [ ]
+        cores = []
 
         for solver in self._solver_list:
             cores.extend(list(solver.unsat_core(extra_constraints=extra_constraints)))
@@ -373,7 +408,7 @@ class CompositeFrontend(ConstrainedFrontend):
         if self._unsat:
             return self.constraints
 
-        new_constraints = [ ]
+        new_constraints = []
 
         l.debug("Simplifying %r with %d solvers", self, len(self._solver_list))
         for s in self._solver_list:
@@ -412,11 +447,22 @@ class CompositeFrontend(ConstrainedFrontend):
         for s in self._solver_list:
             s.timeout = t
 
+    @property
+    def max_memory(self):
+        return self._template_frontend.max_memory
+
+    @max_memory.setter
+    def max_memory(self, val):
+        # this is technically wrong. we cannot enforce a memory limit for a pool shared among multiple solvers
+        self._template_frontend.max_memory = val
+        for s in self._solver_list:
+            s.max_memory = val
+
     @staticmethod
     def _merge_with_ancestor(common_ancestor, merge_conditions):
         merged = common_ancestor.branch()
         merged.add([Or(*merge_conditions)])
-        #import ipdb; ipdb.set_trace()
+        # import ipdb; ipdb.set_trace()
         return True, merged
 
     def merge(self, others, merge_conditions, common_ancestor=None):
@@ -426,7 +472,7 @@ class CompositeFrontend(ConstrainedFrontend):
         l.debug("Merging %s with %d other solvers.", self, len(others))
         merged = self.blank_copy()
         common_solvers = self._shared_solvers(others)
-        common_ids = { id(s) for s in common_solvers }
+        common_ids = {id(s) for s in common_solvers}
         l.debug("... %s common solvers", len(common_solvers))
 
         for s in common_solvers:
@@ -437,10 +483,10 @@ class CompositeFrontend(ConstrainedFrontend):
             for v in s.variables:
                 merged._solvers[v] = s
 
-        noncommon_solvers = [ [ s for s in cs._solver_list if id(s) not in common_ids ] for cs in [self]+others ]
+        noncommon_solvers = [[s for s in cs._solver_list if id(s) not in common_ids] for cs in [self] + others]
 
         l.debug("... merging noncommon solvers")
-        combined_noncommons = [ ]
+        combined_noncommons = []
         for ns in noncommon_solvers:
             l.debug("... %d", len(ns))
             if len(ns) == 0:
@@ -451,16 +497,12 @@ class CompositeFrontend(ConstrainedFrontend):
                 combined_noncommons.append(ns[0].combine(ns[1:]))
 
         if len(combined_noncommons):
-            _, merged_noncommon = combined_noncommons[0].merge(
-                combined_noncommons[1:], merge_conditions
-            )
+            _, merged_noncommon = combined_noncommons[0].merge(combined_noncommons[1:], merge_conditions)
 
             merged._owned_solvers.add(merged_noncommon)
             merged._store_child(merged_noncommon)
 
-        merged.constraints = list(
-            itertools.chain.from_iterable(a.constraints for a in merged._solver_list)
-        )
+        merged.constraints = list(itertools.chain.from_iterable(a.constraints for a in merged._solver_list))
         return True, merged
 
     def combine(self, others):
@@ -471,7 +513,8 @@ class CompositeFrontend(ConstrainedFrontend):
         return combined
 
     def split(self):
-        return [ s.branch() for s in self._solver_list ]
+        return [s.branch() for s in self._solver_list]
+
 
 from ..ast import Base
 from ..ast.bool import Or
